//gputty.cpp
//Copyright (C) 2002  Pierre Pronchery
/* This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/



#include "config.h"
#include <gnome.h>
#include <gtkpp/gtkpp.h>


#define GPUTTY_COPYRIGHT "(c) 2002 Pierre 'khorben' Pronchery"
const char* GPUTTY_AUTHORS[] = { "Pierre 'khorben' Pronchery <khorben@defora.org>", NULL };
#define GPUTTY_DESCRIPTION "A Gnome PuTTY clone.\nIt stores and restores Telnet and SSH sessions."
#define GPUTTY_ICON NULL

//application defaults
#define GPUTTY_CONFIG ".gputty"		//config file place in $HOME

//acceptable defaults
#define SECTION "defaults"
#define XTERM_PATH "/usr/bin/x-terminal-emulator -e"
#define SSH_PATH "/usr/bin/ssh"
#define TELNET_PATH "/usr/bin/telnet"
#define HOSTNAME "localhost"
#define PORT 22
typedef enum _Protocol
{
	TELNET,
	SSH
} Protocol;
#define PROTOCOL SSH


typedef struct _Session
{
	gchar* name;
	gchar* hostname;
	gchar* username;
	gint port;
	Protocol protocol;
} Session;


class GPuTTY: public Window
{
	private:
		//common information
		gchar* _xterm_path;
		gchar* _telnet_path;
		gchar* _ssh_path;

		gint _selected_row;	//current selected session

		gchar* _hostname;
		gchar* _username;
		gint _port;
		Protocol _protocol;

		Config _config;

		//stored sessions
		SList _sessions_list;

		//widgets
		//preferences window
		Frame _fhostname;	//hostname choice
		VBox _vhostname;
		HBox _hhostname;
		VBox _vh1;
		VBox _vh2;
		VBox _vh3;
		Label _lhostname;
		Entry _ehostname;
		Label _lport;		//port choice
		Spin _sport;
		Label _lusername;
		Entry _eusername;
		HBox _hh2;
		Label _lprotocol;
		Optionmenu _oprotocol;
		Frame _fsessions;	//stored sessions
		VBox _vsessions;
		HBox _hs1;
		VBox _vs1;
		VBox _vs2;
		Label _lsessions;
		Entry _esessions;
		Scroll _ssessions;
		CList _sessions;
		Button _load;
		Button _save;
		Button _delete;
		HBox _hbottom;		//bottom
		Button _babout;
		Button _connect;
		Button _cancel;

		About _about;		//about
	public:
		GPuTTY();
		~GPuTTY();

		//useful
		void save();
		void load();
		void connect();

		//callbacks
		gint onClose();
		void onAbout();
		void onConnect();
		void onCancel();

		void onHostnameChange();
		void onUsernameChange();
		void onProtocolChange(Protocol protocol);

		void onLoad();
		void onSave();
		void onDelete();

		void onSelectRow(gint row);
		void onUnselectRow(gint row);
};


//callbacks
gint preferences_close(GtkWidget* widget, GdkEvent* event, GPuTTY* data)
{
	return data->onClose();
}
void preferences_about(GtkWidget* widget, GPuTTY* data)
{
	data->onAbout();
}
void preferences_connect(GtkWidget* widget, GPuTTY* data)
{
	data->onConnect();
}
void preferences_cancel(GtkWidget* widget, GPuTTY* data)
{
	data->onCancel();
}

void preferences_hostname_change(GtkWidget* widget, GPuTTY* data)
{
	data->onHostnameChange();
}
void preferences_username_change(GtkWidget* widget, GPuTTY* data)
{
	data->onUsernameChange();
}
void preferences_protocol_change(GtkWidget* widget, GPuTTY* data)
{
	gint i = GPOINTER_TO_INT(gtk_object_get_user_data(GTK_OBJECT(widget)));
	data->onProtocolChange((Protocol)i);
}

void preferences_load(GtkWidget* widget, GPuTTY* data)
{
	data->onLoad();
}
void preferences_save(GtkWidget* widget, GPuTTY* data)
{
	data->onSave();
}
void preferences_delete(GtkWidget* widget, GPuTTY* data)
{
	data->onDelete();
}
void preferences_select_row(GtkCList *clist, gint row, gint column, GdkEventButton *event, GPuTTY* data)
{
	data->onSelectRow(row);
}
void preferences_unselect_row(GtkCList *clist, gint row, gint column, GdkEventButton *event, GPuTTY* data)
{
	data->onUnselectRow(row);
}


GPuTTY::GPuTTY():
Window(_("GPuTTY Configuration")),
	_config(GPUTTY_CONFIG),
	_fhostname(_("Specify your connection")),
	_vhostname(FALSE, 0),
	_hhostname(FALSE, 0),
	_lhostname(_("Host Name")),
	_lport(_("Port")),
	_sport(1.0, 0, PROTOCOL == TELNET ? 23 : 22, 1, 65535, 1, 100),
	_lusername(_("Username")),
	_lprotocol(_("Protocol")),
	_fsessions(_("Load, save or delete a session")),
	_vsessions(FALSE, 0),
	_lsessions(_("Saved Sessions")),
	_ssessions(GTK_POLICY_NEVER, GTK_POLICY_AUTOMATIC),
	_sessions(1),
	_load(_("Load")),
	_save(_("Save")),
	_delete(_("Delete")),
	_hbottom(FALSE, 0),
	_babout(_("About")),
	_connect(_("Connect")),
	_cancel(_("Cancel")),
	_about(PACKAGE, VERSION, GPUTTY_COPYRIGHT, GPUTTY_AUTHORS, GPUTTY_DESCRIPTION, GPUTTY_ICON)
{
	//essentials
	signalConnect("delete_event", (void*)preferences_close, this);

	_xterm_path = g_strdup(XTERM_PATH);
	_ssh_path = g_strdup(SSH_PATH);
	_telnet_path = g_strdup(TELNET_PATH);
	_selected_row = -1;
	_hostname = g_strdup(HOSTNAME);
	_username = g_strdup(g_get_user_name());
	_port = PORT;
	_protocol = PROTOCOL;

	//widgets
	_vh1.add(&_lhostname);
	_ehostname.signalConnect("changed", (void*)preferences_hostname_change, this);
	_vh1.add(&_ehostname, TRUE, TRUE, 0);
	_hhostname.add(&_vh1, TRUE, TRUE, 0);
	_vh2.add(&_lport);
	_vh2.add(&_sport);
	_hhostname.add(&_vh2);
	_vh3.add(&_lusername);
	_eusername.signalConnect("changed", (void*)preferences_username_change, this);
	_vh3.add(&_eusername);
	_hhostname.add(&_vh3);
	_vhostname.add(&_hhostname);
	_hh2.add(&_lprotocol);
	_oprotocol.add(_("Telnet"), (void*)preferences_protocol_change, this, TELNET);
	_oprotocol.add(_("SSH"), (void*)preferences_protocol_change, this, SSH);
	_hh2.addEnd(&_oprotocol, TRUE, TRUE, 0);
	_vhostname.add(&_hh2);
	_fhostname.add(&_vhostname);
	add(&_fhostname);

	_vsessions.add(&_lsessions);

	_vs1.add(&_esessions);
	_sessions.signalConnect("select_row", (void*)preferences_select_row, this);
	_sessions.signalConnect("unselect_row", (void*)preferences_unselect_row, this);
	_ssessions.add(&_sessions);
	_vs1.add(&_ssessions, TRUE, TRUE, 0);
	_hs1.add(&_vs1, TRUE, TRUE, 0);

	_load.signalConnect("clicked", (void*)preferences_load, this);
	_vs2.add(&_load, FALSE, FALSE, 5);
	_save.signalConnect("clicked", (void*)preferences_save, this);
	_vs2.add(&_save, FALSE, FALSE, 5);
	_delete.signalConnect("clicked", (void*)preferences_delete, this);
	_vs2.add(&_delete, FALSE, FALSE, 5);
	_hs1.addEnd(&_vs2, FALSE, FALSE, 5);
	_vsessions.add(&_hs1, TRUE, TRUE, 0);

	_fsessions.add(&_vsessions);
	add(&_fsessions, TRUE, TRUE, 0);

	_babout.signalConnect("clicked", (void*)preferences_about, this);
	_hbottom.add(&_babout);
	_connect.signalConnect("clicked", (void*)preferences_connect, this);
	_hbottom.addEnd(&_connect);
	_cancel.signalConnect("clicked", (void*)preferences_cancel, this);
	_hbottom.addEnd(&_cancel);
	add(&_hbottom);

	load();

	//fill dialog
	_ehostname.setText(_hostname);
	_sport.setValue(_port);
	_eusername.setText(_username);
	_eusername.sensitive(_protocol == SSH);
	_oprotocol.setHistory(_protocol);

	showAll();
}

GPuTTY::~GPuTTY()
{
	//free gchar*
	g_free(_xterm_path);
	g_free(_ssh_path);
	g_free(_telnet_path);
	g_free(_hostname);
	g_free(_username);

	//free lists
	for(Session* s; (s = (Session*)_sessions_list.get(0)); )
	{
		g_free(s->name);
		g_free(s->hostname);
		g_free(s->username);
		_sessions_list.remove(s);
		delete s;
	}
}


//useful
void GPuTTY::save()
{
	//open file for writing
	g_print(_("Updating config file: "));
	if(_config.open(TRUE) == FALSE)
	{
		g_print(_("failure.\n"));
		return;
	}

	//write common values
	gchar strp[6];
	_config.setVariable(SECTION, "hostname", _hostname);
	_port = _sport.getValueAsInt();
	snprintf(strp, 6, "%d", _port);
	_config.setVariable(SECTION, "port", strp);
	_config.setVariable(SECTION, "username", _username);
	snprintf(strp, 2, "%d", _protocol);
	_config.setVariable(SECTION, "protocol", strp);

	_config.setVariable(SECTION, "xterm_path", _xterm_path);
	_config.setVariable(SECTION, "ssh_path", _ssh_path);
	_config.setVariable(SECTION, "telnet_path", _telnet_path);

	//store sessions
	gint n = strlen("session") + 3;
	gchar str[n];
	Session* s;
	for(gint i = 0; (s = (Session*)_sessions_list.get(i)) && i < 100; i++)
	{
		snprintf(str, n, "session%d", i);
		_config.setVariable(str, "name", s->name);
		_config.setVariable(str, "hostname", s->hostname);
		snprintf(strp, 6, "%d", s->port);
		_config.setVariable(str, "port", strp);
		_config.setVariable(str, "username", s->username);
		snprintf(strp, 2, "%d", s->protocol);
		_config.setVariable(str, "protocol", strp);
	}

	_config.close();
	g_print(_("success.\n"));
}

void GPuTTY::load()
{
	if(_config.open(FALSE) == FALSE)
		return;

	gchar* str;
	_config.getVariable(SECTION, "hostname", &str);
	if(str != NULL) { g_free(_hostname); _hostname = str; }

	_config.getVariable(SECTION, "port", &str);
	if(str != NULL) { _port = atoi(str); g_free(str); }

	_config.getVariable(SECTION, "username", &str);
	if(str != NULL)	{ g_free(_username); _username = str; }

	_config.getVariable(SECTION, "protocol", &str);
	if(str != NULL) { _protocol = (Protocol)atoi(str); g_free(str); }

	_config.getVariable(SECTION, "xterm_path", &str);
	if(str != NULL) { g_free(_xterm_path); _xterm_path = str; }

	_config.getVariable(SECTION, "ssh_path", &str);
	if(str != NULL) { g_free(_ssh_path); _ssh_path = str; }

	_config.getVariable(SECTION, "telnet_path", &str);
	if(str != NULL) { g_free(_telnet_path); _telnet_path = str; }

	gint n = strlen("session") + 3;
	gchar buf[n];
	Session* s;
	for(gint i = 0; i < 100; i++)
	{
		s = new Session;
		snprintf(buf, n, "session%d", i);
		_config.getVariable(buf, "name", &s->name);
		_config.getVariable(buf, "hostname", &s->hostname);
		_config.getVariable(buf, "port", &str);
		if(str != NULL) { s->port = atoi(str); g_free(str); }
		_config.getVariable(buf, "username", &s->username);
		_config.getVariable(buf, "protocol", &str);
		if(str != NULL) { s->protocol = (Protocol)atoi(str); g_free(str); }
		if(s->name != NULL)
		{
			_sessions_list.append(s);
			gint i = _sessions.append(&s->name);
			_sessions.setData(i, s);
		}
		else
		{
			i = 101;
			delete s;
		}
	}

	_config.close();
}

void GPuTTY::connect()
	//when a connection is requested
{
	_port = _sport.getValueAsInt();
	gint n;
	if(_protocol == SSH)
	{
		n = strlen(_xterm_path) + 1 + strlen(_ssh_path) + 4 + 6 + 1 + strlen(_username) + 1 + strlen(_hostname) + 1;
		gchar buf[n];
		snprintf(buf, n, "%s %s -p %d %s@%s", _xterm_path, _ssh_path, _port, _username, _hostname);
		g_print("%s\n", buf);
		gnome_execute_shell(NULL, buf);
	}
	else
	{
		n = strlen(_xterm_path) + 1 + strlen(_telnet_path) + 1 + strlen(_hostname) + 1 + 6 + 1;
		gchar buf[n];
		snprintf(buf, n, "%s %s %s %d", _xterm_path, _telnet_path, _hostname, _port);
		g_print("%s\n", buf);
		gnome_execute_shell(NULL, buf);
	}
}


//callbacks
gint GPuTTY::onClose()
	//when the preferences window is closed
{
	save();
	gtk_main_quit();
	return TRUE;
}

void GPuTTY::onHostnameChange()
	//when updating hostname
{
	g_free(_hostname);
	_hostname = g_strdup(_ehostname.getText());
}

void GPuTTY::onUsernameChange()
	//when updating username
{
	g_free(_username);
	_username = g_strdup(_eusername.getText());
}

void GPuTTY::onProtocolChange(Protocol protocol)
	//when changing protocol
{
	_port = _sport.getValueAsInt();
	if(protocol == SSH)
	{
		_port = 22;
		_sport.setValue(22);
	}
	else if(protocol == TELNET && _protocol == SSH && _port == 22)
	{
		_port = 23;
		_sport.setValue(23);
	}
	_protocol = protocol;
	_eusername.sensitive(protocol == SSH);
}

void GPuTTY::onLoad()
	//when a session is loaded
{
	if(_selected_row < 0)
		return;
	Session* s = (Session*)_sessions_list.get(_selected_row);
	_ehostname.setText(s->hostname);
	_sport.setValue(s->port);
	_eusername.setText(s->username);
	_protocol = s->protocol;
	_eusername.sensitive(_protocol == SSH);
	_oprotocol.setHistory(_protocol);
}

void GPuTTY::onSave()
	//when a session is saved
{
	gchar* name;
	gchar* hostname;
	gchar* username;
	//check if hostname and session are valid FIXME
	name = _esessions.getText();
	hostname = _ehostname.getText();
	_port = _sport.getValueAsInt();
	username = _eusername.getText();

	if(strcmp("", name) == 0)
		return;
	if(strcmp("", hostname) == 0)
		return;
	if(strcmp("", username) == 0)
		username = g_get_user_name();

	//store as a Session
	Session* s = new Session;
	s->name = g_strdup(name);
	s->hostname = g_strdup(hostname);
	s->port = _port;
	s->username = g_strdup(username);
	s->protocol = _protocol;

	//add to sessions list
	_sessions_list.append(s);

	//add to sessions widget
	gint i = _sessions.append(&s->name);
	_sessions.setData(i, s);		//much more practical
}

void GPuTTY::onDelete()
	//when a session is deleted
{
	if(_selected_row < 0)
		return;

	Session* s = (Session*)_sessions_list.get(_selected_row);
	if(s == NULL)
	{
		g_print("Error while deleting session %d\n", _selected_row);
		return;
	}
	_sessions_list.remove(s);
	g_free(s->name);
	g_free(s->hostname);
	g_free(s->username);
	delete s;

	gint n = strlen("session") + 3;
	gchar str[n];
	snprintf(str, n, "session%d", _selected_row);
	_config.deleteSection(str);

	_sessions.remove(_selected_row);

	_selected_row = -1;
}

void GPuTTY::onSelectRow(gint row)
	//when a session is selected
{
	_selected_row = row;
	_esessions.setText(((Session*)_sessions_list.get(row))->name);
}

void GPuTTY::onUnselectRow(gint row)
	//when a session is unselected
{
	_selected_row = -1;
}

void GPuTTY::onAbout()
{
	_about.show();
}

void GPuTTY::onConnect()
{
	connect();
}

void GPuTTY::onCancel()
{
	gtk_main_quit();
}


//main
int main(int argc, char* argv[])
{
	bindtextdomain(PACKAGE, LOCALEDIR);
	textdomain(PACKAGE);

	gnome_init(PACKAGE, VERSION, argc, argv);

	GPuTTY gputty;

	gtk_main();

	return 0;
}
